from logging import NullHandler, getLogger

from .botbase import BotBase, get_handler
from .checks import *
from .emojis import Emojis
from .exceptions import Blacklisted
from .wraps import (
    MyContext,
    MyInter,
    Wrap,
    WrappedChannel,
    WrappedMember,
    WrappedThread,
    WrappedUser,
)
from .models import CogBase

__version__ = "1.17.0"


getLogger(__name__).addHandler(NullHandler())



from __future__ import annotations

from nextcord import Member
from nextcord.ext.commands import MemberConverter

from . import wrap


class WrappedMember(wrap.Wrap, Member):
    @classmethod
    async def convert(cls, ctx, argument: str) -> WrappedMember:
        member: Member = await MemberConverter().convert(ctx=ctx, argument=argument)
        return cls(member, ctx.bot)

    def __getattr__(self, item):
        return getattr(self._wrapped, item)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from asyncpg import Pool


class Blacklist:
    def __init__(self, db: Pool) -> None:
        self.db = db

        self.guilds = set()
        self.users = set()

    def __contains__(self, obj: int) -> bool:
        assert isinstance(obj, int)

        return obj in self.guilds or obj in self.users

    async def add(
        self, obj: int, reason: str = "Unknown reason", guild: bool = False
    ) -> None:
        assert isinstance(obj, int)

        if guild:
            self.guilds.add(obj)
            await self.db.execute(
                """
                INSERT INTO blacklist_guilds (id, reason)
                VALUES ($1, $2)
                ON CONFLICT (id) DO UPDATE
                    SET reason = $2
            """,
                obj,
                reason,
            )
        else:
            self.users.add(obj)
            await self.db.execute(
                """
                INSERT INTO blacklist_users (id, reason)
                VALUES ($1, $2)
                ON CONFLICT (id) DO UPDATE
                    SET reason = $2
            """,
                obj,
                reason,
            )

    async def remove(self, obj: int, guild: bool = False) -> None:
        assert isinstance(obj, int)

        if guild:
            self.guilds.discard(obj)
            await self.db.execute(
                """
                DELETE FROM blacklist_guilds
                WHERE id = $1
            """,
                obj,
            )
        else:
            self.users.discard(obj)
            await self.db.execute(
                """
                DELETE FROM blacklist_users
                WHERE id = $1
            """,
                obj,
            )



# from __future__ import annotations

# from typing import TypeVar, Callable, Union, Coroutine, Any, cast
# from asyncio import iscoroutinefunction
# from functools import wraps

# from nextcord.ext.commands import (
#     BotMissingPermissions,
#     Context,
#     GuildChannelConverter,
#     MissingPermissions,
#     NoPrivateMessage,
#     bot_has_guild_permissions,
#     bot_has_permissions,
#     has_guild_permissions,
#     has_permissions,
#     Command,
# )
# from nextcord import ClientUser, User, Interaction, BaseApplicationCommand

# from .wraps.context import MyContext
# from .wraps.inter import MyInter

# __all__ = (
#     "admin_owner_perms",
#     "bot_admin_perms",
#     "admin_owner_guild_perms",
#     "bot_admin_guild_perms",
#     "admin_owner_or_channel_perms",
#     "bot_admin_or_channel_perms",
#     "admin_owner_or_arg_perms",
#     "bot_admin_or_arg_perms",
# )


# T = TypeVar("T")
# Coro = Coroutine[Any, Any, T]
# CoroFunc = Callable[..., Coro[Any]]
# MaybeCoro = Union[T, Coro[T]]
# CT = TypeVar("CT", bound=Union[Interaction, Context])
# CH = TypeVar("CH", BaseApplicationCommand, Command, CoroFunc)


# # im so sorry if you see this
# def check(predicate) -> Callable[[CH], CH]:
#     def decorator(func: CH) -> CH:
#         if isinstance(func, (BaseApplicationCommand, Command)):
#             func.checks.append(predicate)
#         else:
#             f = cast(CoroFunc, func)
#             if not hasattr(f, "__slash_command_checks__"):
#                 f.__slash_command_checks__ = []

#             if not hasattr(f, "__commands_checks__"):
#                 f.__commands_checks__ = []

#             f.__slash_command_checks__.append(predicate)
#             f.__commands_checks__.append(predicate)

#         return func

#     if iscoroutinefunction(predicate):
#         decorator.predicate = predicate
#     else:

#         @wraps(predicate)
#         async def wrapper(h):
#             return predicate(h)

#         decorator.predicate = wrapper

#     return decorator


# def admin_owner_perms(**perms: bool):
#     """user has admin or is me or permissions"""

#     async def extended_check(ctx: MyContext | MyInter) -> bool:
#         if ctx.guild is None or isinstance(ctx.author, User):
#             raise NoPrivateMessage()

#         a = cast(User, ctx.author)  # so its a user not member idk

#         original = has_permissions(**perms).predicate
#         if (
#             ctx.author.guild_permissions.administrator
#             or await ctx.bot.is_owner(a)
#             or await original(ctx)
#         ):
#             return True
#         else:
#             raise MissingPermissions(list(perms.keys()))

#     return check(extended_check)


# def bot_admin_perms(**perms: bool):
#     """bot is admin or has perms (channel based)"""

#     async def extended_check(ctx: MyContext | MyInter) -> bool:
#         if ctx.guild is None or isinstance(ctx.me, ClientUser):
#             raise NoPrivateMessage()

#         original = bot_has_permissions(**perms).predicate

#         if ctx.me.guild_permissions.administrator or await original(ctx):
#             return True
#         else:
#             raise BotMissingPermissions(list(perms.keys()))

#     return check(extended_check)


# def admin_owner_guild_perms(**perms: bool):
#     """user is admin or me or permissions (guild based)"""

#     async def extended_check(ctx: MyContext | MyInter) -> bool:
#         if ctx.guild is None or isinstance(ctx.author, User):
#             raise NoPrivateMessage()

#         original = has_guild_permissions(**perms).predicate
#         a = cast(User, ctx.author)  # so its a user not member idk

#         if (
#             await ctx.bot.is_owner(a)
#             or ctx.author.guild_permissions.administrator
#             or await original(ctx)
#         ):
#             return True
#         else:
#             raise MissingPermissions(list(perms.keys()))

#     return check(extended_check)


# def bot_admin_guild_perms(**perms: bool):
#     """bot is admin or perms (guild based)"""

#     async def extended_check(ctx: MyContext | MyInter) -> bool:
#         if ctx.guild is None or isinstance(ctx.author, User):
#             raise NoPrivateMessage()

#         original = bot_has_guild_permissions(**perms).predicate
#         a = cast(User, ctx.author)  # so its a user not member idk

#         if (
#             await ctx.bot.is_owner(a)
#             or ctx.author.guild_permissions.administrator
#             or await original(ctx)
#         ):
#             return True
#         else:
#             raise MissingPermissions(list(perms.keys()))

#     return check(extended_check)


# def admin_owner_or_channel_perms(channel: str, **perms: bool):
#     """user is admin or me or has perms on channel"""

#     async def extended_check(ctx: MyContext | MyInter) -> bool:
#         if ctx.guild is None or isinstance(ctx.author, User):
#             raise NoPrivateMessage()

#         c = cast(Context, Context)
#         ch = await GuildChannelConverter().convert(c, channel)
#         permissions = ch.permissions_for(ctx.author)

#         missing = [
#             perm for perm, value in perms.items() if getattr(permissions, perm) != value
#         ]
#         a = cast(User, ctx.author)  # so its a user not member idk

#         if await ctx.bot.is_owner(a) or permissions.administrator or not missing:
#             return True
#         else:
#             raise MissingPermissions(missing)

#     return check(extended_check)


# def bot_admin_or_channel_perms(channel: str, **perms):
#     """bot is admin or has perms on channel"""

#     async def extended_check(ctx: MyContext | MyInter) -> bool:
#         if ctx.guild is None or isinstance(ctx.me, ClientUser):
#             raise NoPrivateMessage()

#         c = cast(Context, Context)
#         ch = await GuildChannelConverter().convert(c, channel)
#         permissions = ch.permissions_for(ctx.me)

#         missing = [
#             perm for perm, value in perms.items() if getattr(permissions, perm) != value
#         ]

#         if permissions.administrator or not missing:
#             return True
#         else:
#             raise BotMissingPermissions(missing)

#     return check(extended_check)


# def admin_owner_or_arg_perms(arg: int, **perms: bool):
#     """user is admin or me or has perms on channel"""

#     async def extended_check(ctx: MyContext | MyInter) -> bool:
#         if ctx.guild is None:
#             raise NoPrivateMessage()

#         ch = ctx.args[arg]
#         permissions = ch.permissions_for(ctx.author)

#         missing = [
#             perm for perm, value in perms.items() if getattr(permissions, perm) != value
#         ]
#         a = cast(User, ctx.author)  # so its a user not member idk

#         if await ctx.bot.is_owner(a) or permissions.administrator or not missing:
#             return True
#         else:
#             raise MissingPermissions(missing)

#     return check(extended_check)


# def bot_admin_or_arg_perms(arg: int, **perms):
#     """bot is admin or has perms on channel"""

#     async def extended_check(ctx: MyContext | MyInter) -> bool:
#         if ctx.guild is None or isinstance(ctx.me, ClientUser):
#             raise NoPrivateMessage()

#         ch = ctx.args[arg]
#         permissions = ch.permissions_for(ctx.me)

#         missing = [
#             perm for perm, value in perms.items() if getattr(permissions, perm) != value
#         ]

#         if permissions.administrator or not missing:
#             return True
#         else:
#             raise BotMissingPermissions(missing)

#     return check(extended_check)
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME: imple# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME
# FIXME: implement for the new slash checks api



from __future__ import annotations

from typing import TYPE_CHECKING

from nextcord.ext.commands import Context

from . import wrap

if TYPE_CHECKING:
    from ..botbase import BotBase


class MyContext(Context, wrap.Wrap):
    bot: BotBase

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._bot = self.bot

        self.message = self.bot.get_wrapped_message(self.message)



from __future__ import annotations

from typing import TYPE_CHECKING

from nextcord import (
    Message,
    Embed,
    WebhookMessage,
    TextChannel,
    Thread,
    DMChannel,
    PartialMessageable,
)
from nextcord.abc import Messageable
from nextcord.utils import utcnow
from nextcord.embeds import _EmptyEmbed

if TYPE_CHECKING:
    from typing import Any

    from ..botbase import BotBase


channels = (TextChannel, PartialMessageable, Thread, DMChannel, Messageable)


class Wrap:
    def __init__(self, wrapped, bot: BotBase):
        self._wrapped = wrapped
        self._bot = bot

        if isinstance(wrapped, type(self)):
            self._wrapped = wrapped._wrapped

    def __instancecheck__(self, instance: Any):
        return isinstance(instance, type(self._wrapped))

    def __subclasscheck__(self, subclass: Any):
        return issubclass(subclass, self._wrapped)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, (type(self._wrapped), type(self))):
            return False

        if isinstance(other, type(self._wrapped)):
            return other.id == self._wrapped.id

        return other._wrapped.id == self._wrapped.id

    def __hash__(self):
        return hash(self._wrapped)

    async def send_author_embed(self, text: str, **kwargs):
        return await self.send_embed(
            author=text,
            cmd_invoker=False,
            invoker_in_author=True,
            try_channel=True,
            **kwargs,
        )

    async def send_embed(
        self,
        title: str | _EmptyEmbed = Embed.Empty,
        desc: str | _EmptyEmbed = Embed.Empty,
        *,
        author: str | None = None,
        image: str | None = None,
        thumbnail: str | None = None,
        color=None,
        target=None,
        reply: bool = False,
        contain_timestamp: bool = True,
        cmd_invoker: bool = True,
        invoker_in_author: bool = False,
        try_channel: bool = False,
        **kwargs,
    ) -> Message | WebhookMessage | None:
        from .context import MyContext
        from .channel import WrappedChannel
        from .inter import MyInter
        from .thread import WrappedThread
        from .member import WrappedMember
        from .user import WrappedUser

        target = target or (
            self.message  # ctx, reply=True
            if reply and isinstance(self, MyContext)
            else self.channel  # ctx, reply=False
            if isinstance(self, MyContext)
            else self  # Anything else (member.send)
        )

        embed = Embed(title=title, description=desc, color=color or self._bot.color)

        if contain_timestamp and isinstance(self, MyContext):
            # Doesnt work on Channels, Users, Members
            embed.timestamp = self.message.created_at
        elif contain_timestamp:
            embed.timestamp = utcnow()

        if cmd_invoker and not isinstance(self, (WrappedChannel, WrappedThread)):
            if isinstance(self, (MyContext)):
                text = self.author.display_name
                icon_url = self.author.display_avatar.url
            elif isinstance(self, MyInter):
                text = self.user.display_name
                icon_url = self.user.display_avatar.url
            elif isinstance(self, (WrappedUser, WrappedMember)):
                text = self.display_name
                icon_url = self.display_avatar.url
            else:
                raise TypeError(f"{type(self).__name__} cannot get invoker")

            embed.set_footer(text=text, icon_url=icon_url)

        if author:
            if isinstance(self, (MyContext)):
                icon_url = self.author.display_avatar.url
            elif isinstance(self, MyInter):
                icon_url = self.user.display_avatar.url
            elif isinstance(self, (WrappedUser, WrappedMember)):
                icon_url = self.display_avatar.url
            else:
                raise TypeError(f"{type(self).__name__} cannot get invoker")

            if invoker_in_author:
                embed.set_author(name=author, icon_url=icon_url)
            else:
                embed.set_author(name=author)

        if image:
            embed.set_image(url=image)
        if thumbnail:
            embed.set_thumbnail(url=thumbnail)

        if reply and isinstance(target, Message):
            return await target.reply(embed=embed, **kwargs)
        elif isinstance(target, Message):
            return await target.channel.send(embed=embed, **kwargs)
        elif try_channel and isinstance(target, MyInter) and target.response.is_done():
            if isinstance(target.channel, channels):
                return await target.channel.send(embed=embed, **kwargs)
            else:
                return await target.send(embed=embed, **kwargs)
        elif isinstance(
            target,
            (
                WrappedUser,
                WrappedMember,
                WrappedChannel,
                WrappedThread,
                MyInter,
            ),
        ):
            return await target.send(embed=embed, **kwargs)
        else:
            raise TypeError(f"{type(self).__name__} cannot send embeds")



from __future__ import annotations

from typing import TYPE_CHECKING

from nextcord import abc, Permissions, DMChannel
from nextcord.ext.commands import TextChannelConverter

from . import wrap

if TYPE_CHECKING:
    from typing import Union


class WrappedChannel(wrap.Wrap, abc.GuildChannel, abc.PrivateChannel):  # type: ignore
    @classmethod
    async def convert(cls, ctx, argument: str) -> WrappedChannel:
        channel: Union[
            abc.GuildChannel, abc.PrivateChannel
        ] = await TextChannelConverter().convert(ctx=ctx, argument=argument)
        return cls(channel, ctx.bot)

    def __getattr__(self, item):
        return getattr(self._wrapped, item)

    def permissions_for(self, *args, **kwargs):
        if not hasattr(self, "guild"):
            return Permissions.all()

        return super().permissions_for(*args, **kwargs)



from __future__ import annotations

from difflib import get_close_matches
from itertools import groupby
from sys import stderr
from traceback import print_exception
from typing import TYPE_CHECKING, cast
from logging import getLogger

from nextcord import ButtonStyle, Embed, Interaction
from nextcord.ext.commands import Cog, HelpCommand
from nextcord.ext.menus import ButtonMenuPages, ListPageSource, PageSource
from nextcord.ui import Select, Button
from nextcord.utils import format_dt
from botbase import MyContext, BotBase, MyInter

if TYPE_CHECKING:
    from nextcord.ext.commands import Command, Group
    from nextcord.ui import Item


log = getLogger(__name__)


class MyMenu(ButtonMenuPages):
    ctx: MyContext
    interaction: MyInter


class MultiSource(ListPageSource):
    def __init__(
        self, multi: Group | Cog, commands: list[Command], *, prefix: str | None = None
    ):
        super().__init__(entries=commands, per_page=5)
        self.multi = multi
        self.prefix = prefix
        self.title = f"{self.multi.qualified_name.capitalize()} Commands"
        self.description = self.multi.description

    def format_page(self, menu: MyMenu, commands: list[Command]) -> Embed:
        ctx = menu.ctx if menu.ctx else menu.interaction
        embed = Embed(
            title=self.title,
            description=f"{self.description}\n{ctx.bot.helpinsert}",
            color=ctx.bot.color,
        )
        for command in commands:
            name = command.qualified_name
            signature = f"{self.prefix}{name} {command.signature}"
            embed.add_field(
                name=signature,
                value=command.short_doc or "N/A",
                inline=False,
            )

        maximum = self.get_max_pages()
        if maximum > 1:
            embed.set_author(
                name=f"Page {menu.current_page + 1}/{maximum} "
                f"({len(self.entries)} commands)"
            )

        embed.set_footer(text=f"Use `{self.prefix}help <command>` for more info.")

        return embed


class FrontPageSource(PageSource):
    def is_paginating(self) -> bool:
        return True

    def get_max_pages(self) -> int:
        return 2

    async def get_page(self, num: int) -> FrontPageSource:
        self.index = num
        return self

    def format_page(self, menu: MyMenu, _) -> Embed:
        bt = (
            menu.ctx.bot
            if menu.ctx
            else menu.interaction.client
            if menu.interaction
            else False
        )
        if not bt:
            raise RuntimeError("Why are we here just to suffer?")

        clean_prefix = menu.ctx.clean_prefix if menu.ctx else "/"

        bot = cast(BotBase, bt)
        assert bot.user is not None

        embed = Embed(
            title=bot.helptitle.format(name=bot.user.name),
            description=bot.helpmsg.format(prefix=clean_prefix, name=bot.user.name),
            color=bot.color,
        )

        for name, value in bot.helpfields.items():
            embed.add_field(
                name=name,
                value=value,
                inline=False,
            )

        created_at = format_dt(bot.user.created_at, "F")
        if self.index == 0:
            embed.add_field(
                name="Who are you?",
                value=bot.helpindex.format(created_at=created_at),
                inline=False,
            )
        elif self.index == 1:
            entries: tuple[tuple[str, str], ...] = (
                ("<arg>", "This means the arg is __**required**__."),
                ("[arg]", "This means the arg is __**optional**__."),
                ("{A|B}", "This means that it can be __**either A or B**__."),
                (
                    "[arg...]",
                    "This means you can have multiple args.\n"
                    "__**You do not type in the brackets!**__",
                ),
            )

            embed.add_field(
                name="How Do I Use You?", value="Here is how the help command works:"
            )

            for name, value in entries:
                embed.add_field(name=name, value=value, inline=False)

        return embed


class HelpSelect(Select["HelpView"]):
    def __init__(
        self, commands: dict[Cog | Group, list[Command]], bot: BotBase
    ) -> None:
        super().__init__(
            placeholder="Select a category...", min_values=1, max_values=1, row=0
        )
        self.commands = commands
        self.bot = bot
        self.fill()

    def fill(self) -> None:
        self.add_option(
            label="Index",
            emoji=self.bot.emojiset.index,
            value="index",
            description="The main page on how to use the bot!",
        )
        for cog, commands in self.commands.items():
            if not commands or cog.qualified_name in ("Jishaku", "help"):
                continue
            description = cog.description or None
            emoji = getattr(cog, "emoji", None)
            self.add_option(
                label=cog.qualified_name.capitalize(),
                emoji=emoji,
                value=cog.qualified_name,
                description=description,
            )

    async def callback(self, interaction: Interaction):
        assert self.view is not None
        value = self.values[0]
        if value == "index":
            await self.view.change_source(FrontPageSource())
            await interaction.response.pong()
        else:
            cog = self.bot.get_cog(value)
            if cog is None:
                await interaction.response.send_message(
                    "Somehow that doesn't exist...", ephemeral=True
                )
                return

            commands = self.commands[cog]
            if not commands:
                await interaction.response.send_message(
                    "That category is empty!", ephemeral=True
                )
                return

            source = MultiSource(cog, commands, prefix=self.view._ctx.clean_prefix)
            await self.view.change_source(source)
            await interaction.response.pong()


class HelpView(ButtonMenuPages):
    def __init__(
        self, source: PageSource, ctx: MyContext, cmds: dict[Cog | Group, list[Command]]
    ) -> None:
        super().__init__(source=source, style=ButtonStyle.blurple)
        if len(cmds) != 1:
            # HACK
            children = self.children.copy()
            for child in children:
                self.remove_item(child)
            self.add_item(HelpSelect(cmds, ctx.bot))
            for child in children:
                self.add_item(child)
            # HACK

        self.ctx = ctx
        self._ctx = ctx

    async def on_error(
        self, error: Exception, item: Item, interaction: Interaction
    ) -> None:
        if interaction.response.is_done():
            await interaction.followup.send(
                "An unknown error occurred, sorry", ephemeral=True
            )
        else:
            await interaction.response.send_message(
                "An unknown error occurred, sorry", ephemeral=True
            )
        print_exception(type(error), error, error.__traceback__, file=stderr)

    async def interaction_check(self, interaction: Interaction) -> bool:
        if interaction.user and interaction.user.id in (
            self._ctx.bot.owner_id,
            self._ctx.author.id,
        ):
            return True
        await interaction.response.send_message(
            f"This menu is for {self._ctx.author.mention}, use {self._ctx.clean_prefix}help to have a menu to yourself.",
            ephemeral=True,
        )
        return False

    async def on_timeout(self):
        for child in self.children:
            if isinstance(child, Button) or isinstance(child, Select):
                child.disabled = True

        if self.message is not None:
            await self.message.edit(view=self)

    async def _set_all_disabled(self, disable: bool):
        for child in self.children:
            if isinstance(child, Button) or isinstance(child, Select):
                child.disabled = disable
        await self._update_view()


class MyHelp(HelpCommand):
    context: MyContext

    def get_command_signature(self, command: Command | Group) -> str:
        name = command.qualified_name
        return f"{self.context.clean_prefix}{name} {command.signature}"

    async def send_bot_help(self, mapping: dict[Cog, Command]) -> None:
        bot = self.context.bot

        if coggie := bot.get_cog("Help"):
            mapping.pop(coggie)

        def key(command) -> str:
            cog = command.cog
            return cog.qualified_name if cog else "\U0010ffff"

        entries: list[Command] = [
            c for c in sorted(bot.commands, key=key) if not c.hidden
        ]

        all_commands: dict[Cog, list[Command]] = {}
        for name, children in groupby(entries, key=key):
            if name == "U0010ffff":
                continue

            cog = bot.get_cog(name)
            if cog is not None:
                all_commands[cog] = sorted(children, key=lambda c: c.qualified_name)

        menu = HelpView(FrontPageSource(), ctx=self.context, cmds=all_commands)  # type: ignore
        if isinstance(self.context, Interaction):
            await menu.start(interaction=self.context)
        else:
            await menu.start(self.context)

    async def send_cog_help(self, cog: Cog):
        entries: list[Command] = [
            c
            for c in sorted(cog.get_commands(), key=lambda c: c.qualified_name)
            if not c.hidden
        ]
        menu = HelpView(
            source=MultiSource(cog, prefix=self.context.clean_prefix, commands=entries),
            cmds={cog: entries},
            ctx=self.context,
        )
        if isinstance(self.context, Interaction):
            await menu.start(interaction=self.context)
        else:
            await menu.start(self.context)

    async def send_command_help(self, command):
        embed = Embed(colour=self.context.bot.color)
        self.common_command_formatting(embed, command)
        await self.context.send(embed=embed)

    async def send_group_help(self, group: Group):
        subcommands = list(group.commands)
        if len(subcommands) == 0:
            return await self.send_command_help(group)

        entries = [c for c in subcommands if not c.hidden]
        if len(entries) == 0:
            return await self.send_command_help(group)

        source = MultiSource(group, entries, prefix=self.context.clean_prefix)
        self.common_command_formatting(source, group)
        menu = HelpView(source, ctx=self.context, cmds={group: entries})
        if isinstance(self.context, Interaction):
            await menu.start(interaction=self.context)
        else:
            await menu.start(self.context)

    async def send_error_message(self, error: Exception) -> None:
        if "command" in self.context.kwargs:
            cmd = self.context.kwargs["command"]
            cmds = [cmd.name for cmd in self.context.bot.commands if not cmd.hidden]
            matches = get_close_matches(cmd, cmds)
            if len(matches) > 0:
                msg = f'Command "`{cmd}`" not found, maybe you meant "`{matches[0]}`"?'
            else:
                msg = f'Command "`{cmd}`" not found, try {self.context.clean_prefix}help to see available commands.'
        else:
            msg = str(error)
        embed = Embed(title="Error", description=msg, color=self.context.bot.color)
        await self.context.send_embed("Error", msg)

    def common_command_formatting(
        self, embed: Embed | MultiSource, command: Command
    ) -> None:
        embed.title = self.get_command_signature(command)
        if command.description:
            embed.description = f"{command.description}\n\n{command.help}"
        else:
            embed.description = command.help or "No help found..."

        if isinstance(embed, Embed):
            if command.aliases:
                embed.add_field(
                    name="Aliases", value=", ".join(command.aliases), inline=False
                )
            if e := command.extras.get("example"):
                embed.add_field(name="Example", value=e, inline=False)


class Help(Cog, name="help", description="Get some help!"):
    def __init__(self, bot: BotBase):
        self.bot = bot
        bot.help_command = MyHelp()
        bot.help_command.cog = self

    # @property
    # def emoji(self) -> str:
    #     return self.bot.emojiset.help


def setup(bot: BotBase):
    bot.add_cog(Help(bot))



class Emojis(dict):
    def __init__(self, d: dict = None) -> None:
        if d:
            for k, v in d.items():
                self.__setattr__(k, v)

    def __getattr__(self, x):
        return self.get(x, None)

    def __getitem__(self, x):
        return self.get(x, None)

    def __setattr__(self, k, v):
        self[k] = v



from sys import argv
from asyncio import run

from asyncpg import create_pool


table = """
CREATE TABLE IF NOT EXISTS blacklist_users (
    id BIGINT PRIMARY KEY,
    reason VARCHAR
);
CREATE TABLE IF NOT EXISTS blacklist_guilds (
    id BIGINT PRIMARY KEY,
    reason VARCHAR
);
CREATE TABLE IF NOT EXISTS guilds (
    id BIGINT PRIMARY KEY,
    prefix VARCHAR
);
CREATE TABLE IF NOT EXISTS commands (
    command VARCHAR NOT NULL,
    guild BIGINT,
    channel BIGINT,
    member BIGINT NOT NULL,
    amount INT NOT NULL,
    UNIQUE(command, guild, channel, member)
);
"""
tables = table.split(";")


async def init(url: str):
    pool = await create_pool(url)
    assert pool is not None

    for table in tables:
        await pool.execute(table)

    print("Done!")
    await pool.close()
    exit(0)


def main():
    if len(argv) == 1:
        print("Usage: botbase <url>")
        exit(1)

    run(init(argv[1]))



from __future__ import annotations

from typing import TYPE_CHECKING

from nextcord import Embed, Object
from nextcord.ext.commands import group, is_owner

from ..models import CogBase
from ..wraps import MyContext

if TYPE_CHECKING:
    from ..botbase import BotBase


class BlacklistCog(CogBase):
    @group(invoke_without_command=True, hidden=True)
    @is_owner()
    async def blacklist(self, ctx: MyContext) -> None:
        await ctx.send_help(ctx.command)

    @blacklist.group(invoke_without_command=True)
    @is_owner()
    async def add(self, ctx: MyContext) -> None:
        await ctx.send_help(ctx.command)

    @add.command(name="person")
    @is_owner()
    async def add_person(self, ctx: MyContext, user: Object, *, reason=None) -> None:
        assert self.bot.blacklist is not None
        await self.bot.blacklist.add(user.id, reason=reason, guild=False)
        await ctx.send_embed(f"I have added <@{user.id}> to the blacklist.")

    @add.command(name="guild")
    @is_owner()
    async def add_guild(self, ctx: MyContext, guild: Object, *, reason=None) -> None:
        assert self.bot.blacklist is not None
        await self.bot.blacklist.add(guild.id, reason=reason, guild=True)
        await ctx.send_embed(f"I have added the guild `{guild.id}` to the blacklist")

    @blacklist.command()
    @is_owner()
    async def list(self, ctx: MyContext) -> None:
        assert self.bot.blacklist is not None
        if self.bot.blacklist.users:
            user_blacklists = "\n".join(f"`{u}`" for u in self.bot.blacklist.users)
        else:
            user_blacklists = "No user's blacklisted."

        if self.bot.blacklist.guilds:
            guild_blacklists = "\n".join(f"`{g}`" for g in self.bot.blacklist.guilds)
        else:
            guild_blacklists = "No guild's blacklisted."

        await ctx.send(
            embed=Embed(
                title="Blacklists",
                description=f"Users:\n{user_blacklists}\n\nGuilds:\n{guild_blacklists}",
            )
        )

    @blacklist.group(invoke_without_command=True)
    @is_owner()
    async def remove(self, ctx: MyContext) -> None:
        await ctx.send_help(ctx.command)

    @remove.command(name="person")
    @is_owner()
    async def remove_person(self, ctx: MyContext, user: Object) -> None:
        assert self.bot.blacklist is not None
        await self.bot.blacklist.remove(user.id, guild=False)
        await ctx.send_embed("I have completed that action for you.")

    @remove.command(name="guild")
    @is_owner()
    async def remove_guild(self, ctx: MyContext, guild: Object) -> None:
        assert self.bot.blacklist is not None
        await self.bot.blacklist.remove(guild.id, guild=True)
        await ctx.send_embed("I have completed that action for you.")


def setup(bot: BotBase):
    bot.add_cog(BlacklistCog(bot))



from .channel import WrappedChannel
from .context import MyContext
from .inter import MyInter
from .member import WrappedMember
from .thread import WrappedThread
from .user import WrappedUser
from .wrap import Wrap

__all__ = (
    "Wrap",
    "MyContext",
    "MyInter",
    "WrappedChannel",
    "WrappedThread",
    "WrappedUser",
    "WrappedMember",
)



from __future__ import annotations

from typing import TYPE_CHECKING

from nextcord import Interaction, InteractionResponse

from . import wrap

if TYPE_CHECKING:
    from nextcord import VoiceProtocol, ClientUser, Member

    from ..botbase import BotBase
    from .user import WrappedUser
    from .member import WrappedMember


class MyInter(wrap.Wrap, Interaction):
    user: WrappedUser | WrappedMember
    author: WrappedUser | WrappedMember
    prefix = "/"
    clean_prefix = "/"

    def __init__(self, wrapped, bot: BotBase):
        self._wrapped = wrapped

        if isinstance(wrapped, type(self)):
            self._wrapped = wrapped._wrapped

    @property
    def _bot(self) -> BotBase:
        return self.client  # type: ignore

    @property
    def bot(self) -> BotBase:
        return self.client  # type: ignore

    @property
    def voice_client(self) -> VoiceProtocol | None:
        return self.guild and self.guild.voice_client

    @property
    def me(self) -> ClientUser | Member:
        return self.guild.me if self.guild is not None else self.bot.user  # type: ignore

    @property
    def command(self) -> str | None:
        if self.data:
            return self.data.get("name")

    @property  # maybe this is why it no worky
    def response(self) -> InteractionResponse:
        return self._wrapped.response

    def __getattr__(self, item):
        return getattr(self._wrapped, item)



from __future__ import annotations

from nextcord import Thread
from nextcord.ext.commands import ThreadConverter

from . import wrap


class WrappedThread(wrap.Wrap, Thread):
    @classmethod
    async def convert(cls, ctx, argument: str) -> WrappedThread:
        _meta: Thread = await ThreadConverter().convert(ctx=ctx, argument=argument)
        return cls(_meta, ctx.bot)

    def __getattr__(self, item):
        return getattr(self._wrapped, item)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['botbase', 'botbase.coggies', 'botbase.wraps']

package_data = \
{'': ['*']}

install_requires = \
['asyncpg>=0.25.0',
 'jishaku==2.4.0',
 'nextcord-ext-menus>=1.5.2,<2.0.0',
 'nextcord>=2.0.0-beta.2',
 'psutil>=5.9.0,<6.0.0']

entry_points = \
{'console_scripts': ['botbase = botbase.cli:main']}

setup_kwargs = {
    'name': 'ooliver-botbase',
    'version': '1.17.1',
    'description': 'A personal nextcord bot base package for bots.',
    'long_description': '# botbase\n\nThis is a botbase project for [nextcord](https://github.com/nextcord/nextcord) Discord Python bots to reduce boilerplate.\n\n## Config values\n\n`db_enabled: bool` default `True`\n\n`db_url: str` either this or name\n\n`db_name: str` either this or url\n\n`db_user: str` default `"ooliver"`\n\n`db_host str` default `"localhost"`\n\n`version: str` default `"0.0.0"`\n\n`aiohttp_enabled: bool` default `True`\n\n`colors: list[int]` default `[0x9966CC]`\n\n`blacklist_enabled: bool` default `True`\n\n`prefix: str | list[str]`\n\n`helpmsg: str` default [`defaulthelpmsg`](https://github.com/ooliver1/botbase/blob/main/botbase/botbase.py#L38-L47)\n\n`helpindex: str` default [`defaulthelpindex`](https://github.com/ooliver1/botbase/blob/main/botbase/botbase.py#L48-L50)\n\n`helptitle: str` default `"Help Me!"`\n\n`helpfields: dict[str, str]` default `{}`\n\n`helpinsert: str` default `""`\n\n`emojiset: Emojis[str, str]` default `Emojis()`\n\n`logchannel: int` default `None`\n\n`guild_ids: list[int]` default `None`\n',
    'author': 'ooliver1',
    'author_email': 'oliverwilkes2006@icloud.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ooliver1/botbase',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from importlib import import_module
from logging import CRITICAL, INFO, Formatter, getLogger
from logging.handlers import RotatingFileHandler
from os import listdir
from os.path import isfile
from pathlib import Path
from random import choice
from sys import modules
from textwrap import dedent
from typing import TYPE_CHECKING

from aiohttp import ClientSession
from asyncpg import create_pool
from nextcord import Embed, Interaction, Member, Thread, User, abc
from nextcord.ext.commands import Bot, when_mentioned_or

from .blacklist import Blacklist
from .emojis import Emojis
from .wraps import (
    MyContext,
    MyInter,
    WrappedChannel,
    WrappedMember,
    WrappedThread,
    WrappedUser,
)

if TYPE_CHECKING:
    from typing import Any, Callable, Optional, Union

    from asyncpg import Pool
    from nextcord import Guild, Message


log = getLogger(__name__)
defaulthelpmsg = """
HI! Welcome to the help page for {name}!

Use `{prefix}help <command>` for more info on a command,
or `{prefix}help <category>` for more info on a category,

Use the dropdown below to select a category.

Have fun!
"""
defaulthelpindex = """
I have been up since {created_at} and I serve for you!
"""


def get_handler():
    h = RotatingFileHandler(
        "./logs/bot/io.log",
        maxBytes=1000000,
        backupCount=5,
        encoding="utf-8",
    )
    h.setFormatter(
        Formatter(
            "%(levelname)-7s %(asctime)s %(filename)12s:%(funcName)-28s: %(message)s",
            datefmt="%H:%M:%S %d/%m/%Y",
        )
    )
    h.namer = lambda name: name.replace(".log", "") + ".log"
    return h


class BotBase(Bot):
    db: Pool
    session: ClientSession
    blacklist: Optional[Blacklist]

    @staticmethod
    def get_config(config_module: str) -> tuple[str, str]:
        file = modules["__main__"].__file__

        if file is None:
            raise RuntimeError("how")

        path = Path(file)

        if path.parts[-1] == "__main__.py":
            mod = path.parts[-2]
            return f"{mod}.{config_module}", mod
        else:
            return config_module, ""

    def __init__(self, *args, config_module: str = "config", **kwargs) -> None:
        pre = kwargs.pop("command_prefix", self.get_pre)
        saf = kwargs.pop("strip_after_prefix", True)
        ca = kwargs.pop("case_insensitive", True)

        super().__init__(
            *args,
            command_prefix=pre,
            strip_after_prefix=saf,
            case_insensitive=ca,
            **kwargs,
        )

        self.prefix: dict[int, list[str]] = {}

        log = getLogger()
        log.handlers = []
        log.setLevel(INFO)
        h = get_handler()

        log.addHandler(h)
        getLogger("asyncio").setLevel(CRITICAL)

        self.loop.set_exception_handler(self.asyncio_handler)

        cfg, mod = self.get_config(config_module.rstrip(".py"))
        config = import_module(cfg)

        self.mod = mod

        self.db_enabled: bool
        self.db_args: tuple[Any, ...]
        self.db_kwargs: dict[str, Any]

        if db_url := getattr(config, "db_url", None):
            self.db_enabled = True
            self.db_args = (db_url,)
            self.db_kwargs = {}
        elif (
            (db_name := getattr(config, "db_name", None))
            and (db_user := getattr(config, "db_user", "ooliver"))
            and (db_host := getattr(config, "db_host", "localhost"))
        ):
            self.db_enabled = True
            self.db_args = ()
            self.db_kwargs = {
                "database": db_name,
                "user": db_user,
            }
        else:
            self.db_enabled = False
            self.db_args = ()
            self.db_kwargs = {}

        if init := getattr(config, "init", None):
            self.db_kwargs["init"] = init

        self.version: str = getattr(config, "version", "0.0.0")
        self.aiohttp_enabled: bool = getattr(config, "aiohttp_enabled", True)
        self.colors: list[int] = getattr(config, "colors", [0x9966CC])
        self.blacklist_enabled: bool = getattr(config, "blacklist_enabled", True)
        self.default_pre: list[str] = getattr(config, "prefix")
        self.helpmsg: str = getattr(config, "helpmsg", defaulthelpmsg)
        self.helpindex: str = getattr(config, "helpindex", defaulthelpindex)
        self.helpfields: dict[str, str] = getattr(config, "helpfields", {})
        self.helptitle: str = getattr(config, "helptitle", "Help Me {name}!")
        self.helpinsert: str = getattr(config, "helpinsert", "")
        self.emojiset: Any = getattr(config, "emojiset", Emojis())
        self.logchannel: int | None = getattr(config, "logchannel", None)
        self.guild_ids: list[int] | None = getattr(config, "guild_ids", None)

        self._single_events: dict[str, Callable] = {
            "on_message": self.get_wrapped_message,
            "on_interaction": self.get_wrapped_interaction,
        }
        self._double_events: dict[str, Callable] = {
            "on_message_edit": lambda before, after: (
                self.get_wrapped_message(before),
                self.get_wrapped_message(after),
            )
        }

        self.load_extension("jishaku")
        self.load_extension("botbase.coggies.help")

        if self.blacklist_enabled:
            self.load_extension("botbase.coggies.blacklist")
            self.blacklist = None

    @property
    def color(self) -> int:
        return choice(self.colors)

    def asyncio_handler(self, _, context: dict) -> None:
        log = getLogger("notasyncio")
        if context["message"] == "Unclosed client session":
            return

        log.error(
            context["message"]
            + "\n"
            + "\n".join(
                f"{k}: {v}"
                for k, v in context.items()
                if k != "message" and k is not None and v is not None and k != "None"
            )
        )

    async def startup(self) -> None:
        if self.db_enabled:
            db = await create_pool(*self.db_args, **self.db_kwargs)
            assert db is not None
            self.db = db

        if self.aiohttp_enabled:
            self.session = ClientSession()

        if self.blacklist_enabled and self.db_enabled:
            self.blacklist = Blacklist(self.db)

    def run(self, *args, **kwargs) -> None:
        self.loop.create_task(self.startup())

        cog_dir = f"{self.mod}/cogs" if self.mod else "./cogs"
        cogs = Path(cog_dir)

        for ext in cogs.glob("**/*.py"):
            log.info("Found file %s", ext)
            if ext.suffix == ".py":
                a = str(ext).replace("/", ".").rstrip(".py")
                log.info("Loading ext %s", a)
                try:
                    self.load_extension(a)
                except Exception as e:
                    log.info("Failed to load %s", a, exc_info=True)
                else:
                    log.info("Loaded ext %s", a)

        super().run(*args, **kwargs)

    async def close(self, *args, **kwargs) -> None:
        if self.aiohttp_enabled:
            await self.session.close()

        await super().close(*args, **kwargs)

    @staticmethod
    async def get_pre(bot: BotBase, message: Message) -> list[str]:
        if not bot.db_enabled:
            return bot.default_pre

        if message.guild is not None:
            try:
                prefix = bot.prefix[message.guild.id]
            except KeyError:
                prefix = [
                    await bot.db.fetchval(
                        "SELECT prefix FROM guilds WHERE id=$1", message.guild.id
                    )
                ]
                if prefix[0] is None:
                    prefix = bot.default_pre
                    bot.prefix[message.guild.id] = prefix
        else:
            prefix = bot.default_pre
        return when_mentioned_or(*prefix)(bot, message)

    async def on_application_command_error(*_):
        ...  # hope an event handler exists

    async def process_commands(self, message: Message) -> None:
        if message.author.bot:
            return

        if self.blacklist and message.author.id in self.blacklist:
            return log.debug("Ignoring blacklisted user %s", message.author.id)
        elif self.blacklist and message.guild and message.guild.id in self.blacklist:
            return log.debug("Ignoring blacklisted guild %s", message.guild.id)

        ctx = await self.get_context(message, cls=MyContext)

        await self.invoke(ctx)

    async def on_interaction(self, interaction: Interaction) -> None:
        i = self.get_wrapped_interaction(interaction)

        if self.blacklist and i.author.id in self.blacklist:
            return log.debug("Ignoring blacklisted user %s", i.author.id)
        elif self.blacklist and i.guild and i.guild.id in self.blacklist:
            return log.debug("Ignoring blacklisted guild %s", i.guild.id)

        await self.process_application_commands(i)

    async def getch_member(self, guild_id: int, member_id: int) -> WrappedMember:
        guild = await self.getch_guild(guild_id)
        member = guild.get_member(member_id)
        if member is not None:
            return WrappedMember(member, bot=self)

        member = await guild.fetch_member(member_id)
        return WrappedMember(member, bot=self)

    async def getch_channel(self, channel_id: int) -> WrappedChannel | WrappedThread:
        channel = self.get_channel(channel_id)
        if channel:
            return self.get_wrapped_channel(channel)

        channel = await self.fetch_channel(channel_id)
        return self.get_wrapped_channel(channel)

    async def getch_guild(self, guild_id: int) -> Guild:
        guild = self.get_guild(guild_id)
        if guild:
            return guild

        guild = await self.fetch_guild(guild_id)
        return guild

    async def getch_user(self, user_id: int) -> WrappedUser:
        user = self.get_user(user_id)
        if user:
            return WrappedUser(user, bot=self)

        user = await self.fetch_user(user_id)
        return WrappedUser(user, self)

    def get_wrapped_channel(
        self,
        channel: Union[abc.GuildChannel, abc.PrivateChannel, Thread],
    ) -> Union[WrappedThread, WrappedChannel]:
        if isinstance(channel, Thread):
            return WrappedThread(channel, self)

        return WrappedChannel(channel, self)

    def get_wrapped_person(
        self, person: Union[User, Member]
    ) -> Union[WrappedUser, WrappedMember]:
        if isinstance(person, Member):
            return WrappedMember(person, self)

        return WrappedUser(person, self)

    def get_wrapped_message(self, message: Message) -> Message:
        message.channel = self.get_wrapped_channel(message.channel)  # type: ignore
        message.author = self.get_wrapped_person(message.author)

        return message

    def get_wrapped_interaction(self, inter: Interaction) -> MyInter:
        i = MyInter(inter, self)
        if i.user:
            i.user = self.get_wrapped_person(i.user)

        i.author = i.user

        return i

    def dispatch(self, event_name: str, *args: Any, **kwargs: Any) -> None:
        _name = f"on_{event_name}"
        if _name in self._single_events:
            wrapped_arg = self._single_events[_name](args[0])
            super().dispatch(event_name, wrapped_arg)

        elif _name in self._double_events:
            wrapped_first_arg, wrapped_second_arg = self._double_events[_name](
                args[0], args[1]
            )
            super().dispatch(event_name, wrapped_first_arg, wrapped_second_arg)

        else:
            super().dispatch(event_name, *args, **kwargs)

    async def on_command_completion(self, ctx: MyContext):
        if not self.db_enabled:
            return

        if ctx.guild is not None:
            await self.db.execute(
                """INSERT INTO commands (command, guild, channel, member, amount) 
                VALUES ($1,$2,$3,$4,$5) 
                ON CONFLICT (command, guild, channel, member) DO UPDATE
                    SET amount = commands.amount + 1""",
                ctx.command.qualified_name,  # type: ignore
                ctx.guild.id,
                ctx.channel.id,
                ctx.author.id,
                1,
            )
        else:
            a = await self.db.fetchval(
                "SELECT * FROM commands WHERE member=$1 AND channel IS NULL and guild IS NULL",
                ctx.author.id,
            )
            if a is None:
                await self.db.execute(
                    """INSERT INTO commands (command, guild, channel, member, amount) 
                    VALUES ($1,$2,$3,$4,$5)""",
                    ctx.command.qualified_name,  # type: ignore
                    None,
                    None,
                    ctx.author.id,
                    1,
                )
            else:
                await self.db.execute(
                    """UPDATE commands SET amount = commands.amount + 1 
                    WHERE member=$1 AND channel IS NULL and guild IS NULL""",
                    ctx.author.id,
                )

    async def on_guild_join(self, guild: Guild):
        if not self.logchannel:
            return
        elif not self.db_enabled:
            return

        if self.blacklist and guild.id in self.blacklist.guilds:
            log.info("Leaving blacklisted Guild(id=%s)", guild.id)
            await guild.leave()
            return

        assert guild.owner_id is not None

        embed = Embed(
            title="New Guild!",
            description=dedent(
                f"""

        **{guild.name}**
        id: `{guild.id}`
        members: `{guild.member_count}`
        owner: `{guild.owner or await self.fetch_user(guild.owner_id)}`

        """
            ),
            color=self.color,
        )
        await self.get_channel(self.logchannel).send(embed=embed)  # type: ignore

    async def on_guild_remove(self, guild: Guild):
        if guild.unavailable:
            return

        if not self.logchannel:
            return
        elif not self.db_enabled:
            return

        assert guild.owner_id is not None

        embed = Embed(
            title="Removed Guild :(",
            description=dedent(
                f"""

        **{guild.name}**
        id: `{guild.id}`
        members: `{guild.member_count}`
        owner: `{guild.owner or await self.fetch_user(guild.owner_id)}`

        """
            ),
            color=self.color,
        )
        try:
            await self.get_channel(self.logchannel).send(embed=embed)  # type: ignore
        except AttributeError:
            pass



from __future__ import annotations

from nextcord import User
from nextcord.ext.commands import UserConverter

from . import wrap


class WrappedUser(wrap.Wrap, User):
    @classmethod
    async def convert(cls, ctx, argument: str) -> WrappedUser:
        user: User = await UserConverter().convert(
            ctx=ctx, argument=argument
        )
        return cls(user, ctx.bot)

    def __getattr__(self, item):
        return getattr(self._wrapped, item)



from typing import Generic, TypeVar

from nextcord.ext.commands import Cog

from .botbase import BotBase

B = TypeVar("B", bound=BotBase)
__all__: tuple[str, ...] = ("CogBase",)


class CogBase(Cog, Generic[B]):
    def __init__(self, bot: B):
        self.bot = bot
